#include <uade-os-unix.h>
#include <uadeconfig.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include <libgen.h>
#include <ctype.h>

static int sharedmem_length = 65536;

static int os_debug = 0;

static int unix_soundfd;
static int unix_usingoutpipe = 0;

void uade_create_signalhandler(void (*sighandler)(int), int sigtype) {
  int signum;
  struct sigaction *sigact;
  switch(sigtype) {
  case UADE_SIGHUP:
    signum = SIGHUP;
    break;
  case UADE_SIGINT:
    signum = SIGINT;
    break;
  case UADE_SIGKILL:
    signum = SIGKILL;
    break;
  default:
    fprintf(stderr,"%s/uade: uade_create_sighandler: no such signal\n", UADE_MODULE);
    return;
  }
  sigact = calloc(1, sizeof(struct sigaction));
  sigact->sa_handler = sighandler;
  sigaction(signum,sigact,0);
}


void uade_create_var_pid(void) {
  /* ADDED BY XIGH */
  int pid, fd;
  char pidstr[10];
  char varname[256];
  strcpy(varname,"/var/run/uade.");
  strcat(varname,getenv("USER"));
  strcat(varname,".pid");
  pid = getpid();
  sprintf(pidstr, "%d", pid);
  fd = open(varname, O_RDWR|O_CREAT|O_TRUNC, 0644);
  if (fd < 0) {
    fprintf(stderr, "could not create %s\n",varname);
  } else {
    write(fd, pidstr, strlen(pidstr));
    close(fd);
  }
}


void uade_create_watchdog(uade_msgstruct *uade_struct, char *mapname) {
  pid_t father = getpid();
  if(!fork()) {
    while(1) {
      uade_sleep(2);
      if(uade_struct->masterpid) {
	if(kill(uade_struct->masterpid, 0) < 0) break;
      }
    }
    /* remove the shared memory file that is used by uade and xmms plugin */
    (void) remove(mapname);
    kill(father,SIGKILL);
    kill(getpid(),SIGKILL);
  }
}


/* first element of args is the executable file */
/* last element of args table has to be a NULL pointer */
uadepid_t uade_fork_exec(char **args) {
  pid_t slavepid = fork();
  if(slavepid == 0) {
    execv(args[0], args);
    fprintf(stderr, "%s/uade: OH SHIT! execv() has returned!\n", UADE_MODULE);
    fprintf(stderr, "%s/uade: report this bug\n", UADE_MODULE);
    fprintf(stderr, "%s/uade: DON'T try to play with UADE now\n", UADE_MODULE);
    abort();
  } else if(slavepid < 0) {
    fprintf(stderr, "%s/uade: fork error\n", UADE_MODULE);
    slavepid = 0;
  }
  return slavepid;
}


char *uade_get_basename(char *path) {
  return basename(path);
}


FILE *uade_get_file(int item) {
  char post[256];
  char glob[256];
  char temp[256];
  FILE *testfile;
  post[0] = glob[0] = 0;
  switch(item) {
  case UADE_PATH_SCORE:
    strcpy(post, "score");
    break;
  case UADE_PATH_UAERC:
    strcpy(post, "uaerc");
    break;
  case UADE_PATH_UADE:
    strcpy(post, "uade");
    strcpy(glob, UADECONFIG_UADE);
    break;
  case UADE_PATH_FORMATSFILE:
    strcpy(post, "players/uadeformats");
    strcpy(glob, UADECONFIG_FORMATSNAME);
    break;
  default:
    fprintf(stderr, "%s/uade: failed to get path of %d\n", UADE_MODULE, item);
    return 0;
  }
  if(post[0] != 0) {
    testfile = fopen(post, "r");
    if(testfile) {
      if(os_debug)
	fprintf(stderr, "get file: success: path = %s\n", post);
      return testfile;
    }
    sprintf(temp, "%s/.xmms/uade/%s", getenv("HOME"), post);
    testfile = fopen(temp, "r");
    if(testfile) {
      if(os_debug)
	fprintf(stderr, "get file: success: path = %s\n", temp);
      return testfile;
    }
  }
  if(glob[0] != 0) {
    testfile = fopen(glob, "r");
    if(testfile) {
      if(os_debug)
	fprintf(stderr, "get file: success: path = %s\n", glob);
      return testfile;
    }
  }
  fprintf(stderr, "%s/uade: failed to get path of %s\n", UADE_MODULE, post);
  return 0;
}


int uade_get_path(char *path, int item) {
  char post[256];
  char glob[256];
  char temp[256];
  FILE *testfile;
  DIR *testdir;
  post[0] = glob[0] = 0;
  path[0] = 0;
  switch(item) {
  case UADE_PATH_SCORE:
    strcpy(post, "score");
    strcpy(glob, UADECONFIG_SCORENAME);
    break;
  case UADE_PATH_UAERC:
    strcpy(post, "uaerc");
    strcpy(glob, UADECONFIG_UAERC);
    break;
  case UADE_PATH_UADE:
    strcpy(post, "uade");
    testdir = opendir(post);
    if(testdir) {
      closedir(testdir);
      post[0] = 0;
    }
    strcpy(glob, UADECONFIG_UADE);
    break;
  case UADE_PATH_PLAYERDIR:
    strcpy(glob, "players/");
    testdir = opendir(glob);
    if(testdir) {
      closedir(testdir);
      strcpy(path, glob);
      strcat(glob, "uadeformats");
      testfile = fopen(glob, "r");
      if(testfile) {
	fclose(testfile); testfile = 0;
	if(os_debug)
	  fprintf(stderr, "success: path = %s\n", path);
	return 1;
      }
    }
    sprintf(glob, "%s/.xmms/uade/players/", getenv("HOME"));
    testdir = opendir(glob);
    if(testdir) {
      closedir(testdir);
      strcpy(path, glob);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
    sprintf(glob, "%splayers/", UADECONFIG_DATADIR);
    testdir = opendir(glob);
    if(testdir) {
      closedir(testdir);
      strcpy(path, glob);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
    fprintf(stderr, "%s/uade: couldn't get uade playerdir path\n", UADE_MODULE);
    return 0;
  case UADE_PATH_FORMATSFILE:
    strcpy(post, "players/uadeformats");
    strcpy(glob, UADECONFIG_FORMATSNAME);
    break;
  default:
    fprintf(stderr, "%s/uade: failed to get path of %d\n", UADE_MODULE, item);
    return 0;
  }
  if(post[0] != 0) {
    testfile = fopen(post, "r");
    if(testfile) {
      fclose(testfile); testfile = 0;
      strcpy(path, post);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
    sprintf(temp, "%s/.xmms/uade/%s", getenv("HOME"), post);
    testfile = fopen(temp, "r");
    if(testfile) {
      fclose(testfile); testfile = 0;
      strcpy(path, temp);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
  }
  if(glob[0] != 0) {
    testfile = fopen(glob, "r");
    if(os_debug)
      fprintf(stderr, "getpath: trying glob = %s\n", glob);
    if(testfile) {
      fclose(testfile); testfile = 0;
      strcpy(path, glob);
      if(os_debug)
	fprintf(stderr, "success: path = %s\n", path);
      return 1;
    }
  }
  fprintf(stderr,"%s/uade: failed to get path of %s\n", UADE_MODULE, post);
  return 0;
}


int uade_get_temp_name(char *tempname) {
  int fd;
  char dir[256];
  sprintf(dir,"/tmp/%s.uade",getenv("USER"));
  if(mkdir(dir,S_IRUSR | S_IWUSR | S_IXUSR)) {
    if(chmod(dir, S_IRUSR | S_IWUSR | S_IXUSR)) {
      fprintf(stderr,"uade: couldn't create directory for tmpfile\n");
      tempname[0] = 0;
      return 0;
    }
  }
  sprintf(tempname,"%s/mmap.%d.XXXXXX",dir,getpid());
  fd = mkstemp(tempname);
  if(fd<0) {
    fprintf(stderr,"uade warning: couldn't generate tmp filename!\n");
    tempname[0] = 0;
    return 0;
  }
  close(fd);
  remove(tempname);
  return 1;
}


uadepid_t uade_getpid(void) {
  return getpid();
}


int uade_if_normal_file(const char *filename) {
  struct stat s;
  if(stat(filename, &s)==0) {
    if(S_ISREG(s.st_mode) || S_ISLNK(s.st_mode)) {
      return 1;
    }
  }
  return 0;
}


int uade_init_mmap_file(const char *mapfilename) {
  char zerobuf[256];
  FILE *sharedmemFILE;
  int i;
  if(!(sharedmemFILE = fopen(mapfilename,"w+"))) {
    fprintf(stderr,"%s/uade: can not create mmap file (fopen)\n",UADE_MODULE);
    return 0;
  }
  memset(zerobuf,0,256);
  for(i=0;i<sharedmem_length;i+=256) fwrite(zerobuf,1,256,sharedmemFILE);
  fclose(sharedmemFILE);
  return 1;
}


FILE *uade_open_amiga_file(char *filename) {
  FILE *file;
  char *separator;
  char *name;     /* points to file without path */
  char path[256]; /* contains dirpath to the file */
  DIR *dir;
  struct dirent *direntry;
  char loadname[256];

  if ((file = fopen(filename,"rb"))) {
    return file;
  }

  /* case: filename = VOLUME:PATH1[/FILE] */
  if((separator = strchr(filename, (int) ':'))) {
    uade_get_path(path, UADE_PATH_PLAYERDIR);
    if((name = strrchr(separator + 1, (int) '/'))) {
      *name = 0;
      strcat(path, filename);
      if((separator = strstr(path, "Env:"))) {
	strncpy(separator, "ENV:", 4);
      } else if((separator = strstr(path, "env:"))) {
	strncpy(separator, "ENV:", 4);
      }
      strcat(path, "/");
      *name = '/';
      name++;
    } else {
      name = separator + 1;
    }
    // fprintf(stderr, "uae/uade.c: using load path = %s name = %s\n", path, name);
    goto jumpover;
  }

  if((separator = strrchr(filename,(int) '/'))) {
    name = separator + 1;
    *separator = 0;
    path[255] = 0;
    strncpy(path, filename, 255);
    strncat(path, "/", 255);
    *separator = '/';
  } else {
    name = filename;
    strcpy(path,"./");
  }

 jumpover:

  if((dir = opendir(path))==0) {
    path[strlen(path)-1] = 0;
    if((separator = strrchr(path, (int) '/'))) {
      if(isupper((int) separator[1])) {
	separator[1] = (char) tolower((int) separator[1]);
      } else if(islower((int) separator[1])) {
	separator[1] = (char) toupper((int) separator[1]);
      }
      strcat(path, "/");
      if((dir = opendir(path))==0) {
	fprintf(stderr, "uae/uade.c: DTP_LoadFile: Can't open module directory '%s' for scanning!\n", path);
	return 0;
      }
    } else {
      fprintf(stderr, "uae/uade.c: DTP_LoadFile: Can't open module directory '%s' for scanning!\n", path);
      return 0;
    }
  }

  // fprintf(stderr,"name = '%s' path = '%s'\n",name,path);
  while((direntry = readdir(dir))) {
    // fprintf(stderr,"dirent: %s\n",direntry->d_name);
    if(strcasecmp(name, direntry->d_name)==0) {
      loadname[255]=0;
      strncpy(loadname, path, 255);
      strncat(loadname, direntry->d_name, 255);
      if((file = fopen(loadname, "rb"))) break;
    }
  }
  closedir(dir);

  if(!file) {
    fprintf(stderr, "uade: DTP_LoadFile: couldn't load file '%s'\n", filename);
  }

  return file;
}


void uade_init_outpipe(char *parameter) {
  unix_usingoutpipe = 1;
  unix_soundfd = atoi(parameter);
}


void *uade_mmap_file(const char *filename) {
  void *mmapptr;
  int fd;
  fd = open(filename,O_RDWR);
  if(fd < 0) {
    fprintf(stderr,"%s/uade: can not open sharedmem file!\n", UADE_MODULE);
    return 0;
  }
  mmapptr = mmap(0,sharedmem_length,PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
  if(mmapptr == MAP_FAILED) {
    fprintf(stderr,"%s/uade: can not mmap sharedmem file!\n", UADE_MODULE);
    return 0;
  }
  return mmapptr;
}


int uade_send_signal(uadepid_t pid, int sigtype) {
  int signum;
  switch(sigtype) {
  case UADE_SIGALIVE:
    signum = 0;
    break;
  case UADE_SIGHUP:
    signum = SIGHUP;
    break;
  case UADE_SIGINT:
    signum = SIGINT;
    break;
  case UADE_SIGKILL:
    signum = SIGKILL;
    break;
  default:
    fprintf(stderr,"%s/uade: uade_send_signal: unknown signal %d\n", UADE_MODULE, sigtype);
    return -1;
  }
  if(pid > 0)
    return kill(pid, signum);
  else
    return -1;
}


void uade_sleep(int seconds) {
  sleep(seconds);
}


void uade_usleep(int microseconds) {
  usleep(microseconds);
}


/* returns number of milliseconds from the last call
   exception: on the first time this returns zero    */
int uade_timer_poll(void) {
  static struct timeval otv = {0,0};
  struct timeval tv;
  int ret = 0;
  gettimeofday(&tv, 0);
  if(otv.tv_sec != 0 || otv.tv_usec != 0) {
    ret = (tv.tv_sec-otv.tv_sec)*1000 + (tv.tv_usec-otv.tv_usec)/1000;
  }
  otv = tv;
  return ret;
}

char *uade_version(void) {
  return strdup(UADEVERSIONSTRING);
}

void uade_write_to_outpipe(void *ptr, int size) {
  if(unix_usingoutpipe) {
    write(unix_soundfd, ptr, size);
  }
}
