/* XMMS UADE plugin
 *
 * Copyright (C) 2000-2001  Heikki Orsila
 *                          heikki.orsila@ee.tut.fi
 *                          http://www.ee.tut.fi/~heikki
 *
 * This plugin is based on xmms 0.9.6 wavplayer input plugin code. Since
 * then all code has been rewritten.
 *
 * the intital gui code was based onthe null-plugin by
 * Hvard Kvlen. Since then the gui code has evolved and was also mostly
 * rewritten.
 * Formatseditor inspired by nscache 0.3's mimetype editor by Stefan Ondrejicka
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <xmms/plugin.h>
#include <xmms/configfile.h>
#include <xmms/util.h>

#include "uade.h"
#include "uade-os-unix.h"
#include "gui.h"

#include "../src/decrunch/decrunch.h"

static int find_tag (unsigned char *buf, char *tag, int startoffset, int buflen);
static unsigned int readbig (char *ptr);

static void process_ptk_mod (char *credits, unsigned char *buf, int modfilelen);
static void process_aon_mod (char *credits, unsigned char *buf, int modfilelen);
static void process_dmu_mod (char *credits, unsigned char *buf, int modfilelen);
static void process_dm2_mod (char *credits, unsigned char *buf, int modfilelen);


void processmodule(char *credits, char *filename) {
    
    FILE *modfile;
    struct stat st;
    int	modfilelen;
    unsigned char *buf;
    char pre[256]="\0";
    char post[256]="\0";
    
    modfile = fopen(filename, "r");

    if (!modfile){
     return;
    } else {
    if (decrunch (&modfile,filename) <0){
	fprintf (stderr, "decrunching error"); 
	fclose	(modfile);
	return;
        }
    }
    
    fstat ( fileno (modfile), &st);
    modfilelen = st.st_size;

    buf = malloc (modfilelen);
    if (!buf) {
	fprintf(stderr, "can't allocate mem");
	return;
	}
    fread (buf, modfilelen, 1 , modfile);
    fclose (modfile);


    filemagic(buf, pre, post); /*get filetype*/

    if (strcasecmp(pre, "DM2\0")==0) {
		process_dm2_mod(credits, buf ,modfilelen); /*DM2*/

    } else if ((strcasecmp(pre, "MOD\0")==0) ||
	       (strcasecmp(pre, "ADSC\0")==0)  ){
		process_ptk_mod(credits, buf ,modfilelen); /*MOD*/

    } else if ((strcasecmp(pre, "AON4\0")==0) ||
	      (strcasecmp(pre, "AON8\0")==0)){
		process_aon_mod(credits, buf ,modfilelen); /*AON*/

    } else if ((strcasecmp(pre, "DMU\0")==0) ||
	      (strcasecmp(pre, "DMU2\0")==0)){
		process_dmu_mod(credits, buf ,modfilelen); /*DMU*/

    } else {
								/* UNKN*/
        sprintf (credits, "module tile:  %s", get_modulename());
        sprintf (credits, "%s\nmodule type:  %s",credits,  get_playername());
        sprintf (credits, "%s\nfile len:  %d bytes", credits, modfilelen);
	sprintf (credits, "%s\nsubsong range:  min:%d  max:%d\n",credits, get_min_subsong(), get_max_subsong());
	strcat (credits, "\n\n(no further module info available, yet)"); 
    }
    free (buf);
    return;
}

/* Get the info out of the protracker module data*/
static void process_ptk_mod (char *credits, unsigned char *buf, int modfilelen)
{
    int i;

    strcat (credits, "module title:  ");
    strncat (credits, buf, 20);

    sprintf (credits, "%s\nmodule type:  %s",credits,  get_playername());
    sprintf (credits, "%s\nfile len:  %d bytes", credits, modfilelen);

    strcat (credits, "\nmagic ID:  ");
    strncat (credits, buf+0x438,4);

    sprintf (credits, "%s\nmax positions:  %d",credits, buf[0x3b6]);
    sprintf (credits, "%s\nsubsong range:  min:%d  max:%d\n",credits, get_min_subsong(), get_max_subsong());
    
    for (i=0; i < 31; i++){
        if (i < 10) {
    	    sprintf (credits, "%s\ninstr #0%d:  ",credits,i);

	} else {
	    sprintf (credits, "%s\ninstr #%d:  ",credits,i);
    	}
        strncat (credits, buf+0x14+(i*0x1e), 22);
    }

 return;
}

/* Get the info out of the Art of Noise module data*/
static void process_aon_mod (char *credits, unsigned char *buf, int modfilelen)
{
    int offset;
    unsigned int len;

    offset=find_tag (buf, "NAME", 0, modfilelen);
    if (offset != -1) {

	len = readbig(buf+offset+4);
	if (len != 0) {
	strcat (credits, "module title:  ");
	strncat (credits, buf+offset+8, len);
	}
    }

    sprintf (credits, "%s\nmodule type:  %s", credits,get_playername());
    sprintf (credits, "%s\nfile len:  %d bytes", credits, modfilelen);


    offset=find_tag (buf, "AUTH", 0, modfilelen);
    if (offset != -1) {

	len = readbig(buf+offset+4);
	if (len != 0) {
	strcat (credits, "\nauthor/composer:  ");
	strncat (credits, buf+offset+8, len);
	}
    }

    offset=find_tag (buf, "DATE", 0, modfilelen);
    if (offset != -1) {

	len = readbig(buf+offset+4);
	if (len != 0) {
	strcat (credits, "\ncreation date:  ");
	strncat (credits, buf+offset+8, len);
	}
    }

    offset=find_tag (buf, "RMRK", 0, modfilelen);
    if (offset != -1) {
	len = readbig(buf+offset+4);
	if (len != 0) {
	strcat (credits, "\nremarks:  ");
	strncat (credits, buf+offset+8, len);
	}
    }

return;
}

/* Get the info out of the Mugician 1&2 module data*/
static void process_dmu_mod (char *credits, unsigned char *buf, int modfilelen)
{
    int i, maxsong;

    maxsong=get_max_subsong();

    sprintf (credits, "module tile:  %s", get_modulename());
    sprintf (credits, "%s\nmodule type:  %s",credits,  get_playername());
    sprintf (credits, "%s\nfile len:  %d bytes", credits, modfilelen);
    sprintf (credits, "%s\nsubsong range:  min:%d  max:%d\n",credits, get_min_subsong(), maxsong);

    strcat (credits, "\nsongtitles:  ");

    for (i=0; i <= maxsong; i++){
	sprintf (credits, "%s\nsubsong #%d:  ",credits,i);
        strncat (credits, buf+0x50+(i*0x10), 12);
    }


}

/* Get the info out of the Deltamusic 2 module data*/
static void process_dm2_mod (char *credits, unsigned char *buf, int modfilelen)
{
    sprintf (credits, "module tile:  %s", get_modulename());
    sprintf (credits, "%s\nmodule type:  %s",credits,  get_playername());
    sprintf (credits, "%s\nfile len:  %d bytes", credits, modfilelen);

    strcat (credits, "\n\nremarks:  ");
    sprintf (credits, "%s\n%s", credits, buf+0x148);

}

static int find_tag (unsigned char *buf, char *tag, int startoffset, int buflen)
{
    int i;

    if (startoffset > buflen-4) return -1;
	
    for (i=startoffset; i < buflen-3; i++) {
        if (buf[i] == tag[0] && buf[i+1] == tag[1] &&
	    buf[i+2] == tag[2] && buf[i+3] == tag[3]) {
	 return i;    
	}
    }
 return -1;
}

static unsigned int readbig (char *ptr){
    unsigned char *p = (unsigned char *) ptr;
    unsigned int x = p[3]+(p[2]<<8)+(p[1]<<16)+(p[0]<<24);
    return x;
}
