 /* 
  * UAE - The Un*x Amiga Emulator
  * 
  * Support for Linux/USS sound
  * 
  * Copyright 1997 Bernd Schmidt
  */

#include <unistd.h>

#include "uade.h"
#include "uade-os-unix.h"

extern int sound_fd;
extern uae_u16 sndbuffer[];
extern uae_u16 *sndbufpt;
extern int sndbufsize;

static __inline__ int free_in_buffer(void) {
  int fs;
  /* Assumption is that write is always ahead or at the same position
     (this assumption means that if r=w then there is 32k free)
    cases are:
     1: r=w: free = 0 (else case)
     2: w>r: free = r + 32k - w - 1 (else case) (consistent with case 1)
     3: w<r: free = r - w - 1 (first case in if structure)
  */
  if(uade_struct->sbuf_writeoffset < uade_struct->sbuf_readoffset) {
    /* case 3 */
    fs = uade_struct->sbuf_readoffset - uade_struct->sbuf_writeoffset - 1;
  } else {
    /* case 1 and 2 */
    fs = 32768+uade_struct->sbuf_readoffset-uade_struct->sbuf_writeoffset - 1;
  }
  return fs;
}

static __inline__ void check_sound_buffers (void) {
  int fs;
  if ((char *)sndbufpt - (char *)sndbuffer >= sndbufsize) {

    if(uade_mixpar) {
      short * sm = (short*) sndbuffer;
      short * esm = (short*) sndbufpt;
      int mixpar = uade_mixpar;
      int l,r,m;
      for(;sm<esm;) {
	l = (int) sm[0];
	r = (int) sm[1];
	m = (r-l) * mixpar;
	sm[0] = (short) ( ((l<<8) + m) >> 8 );
	sm[1] = (short) ( ((r<<8) - m) >> 8 );
	sm += 2;
      }
    }

    if(uade_slavemode) {
      if(uade_struct->dontwritebit == 0) {
	while(uade_struct->plugin_pause_boolean) usleep(10000);
	fs = free_in_buffer();
	if(fs < sndbufsize) {
	  int sleepcounter;
	  for(sleepcounter=0;sleepcounter<1000000;) {
	    fs = free_in_buffer();
	    if(fs >= sndbufsize) break;
	    if(uade_struct->dontwritebit) break;
	    usleep(10000);
	    sleepcounter+=10000;
	  }
	}
	/* copy sndbuffer to uade shared memory sound buffer (take into
	   account that the buffer is circular) */
	if((uade_struct->sbuf_writeoffset + sndbufsize) > 32768) {
	  int firstsize = 32768-uade_struct->sbuf_writeoffset;
	  memcpy(uade_struct->soundbuffer+uade_struct->sbuf_writeoffset,sndbuffer,firstsize);
	  memcpy(uade_struct->soundbuffer,sndbuffer,sndbufsize-firstsize);
	} else {
	  memcpy(uade_struct->soundbuffer+uade_struct->sbuf_writeoffset,sndbuffer,sndbufsize);
	}
	/* increase write counter */
	uade_struct->sbuf_writeoffset = (uade_struct->sbuf_writeoffset + sndbufsize) % 32768;
      }
    } else if (uade_usingoutpipe) {
      uade_write_to_outpipe(sndbuffer, sndbufsize);
    } else {
      write (sound_fd, sndbuffer, sndbufsize);
    }
    sndbufpt = sndbuffer;
  }
}

#define PUT_SOUND_BYTE(b) do { *(uae_u8 *)sndbufpt = b; sndbufpt = (uae_u16 *)(((uae_u8 *)sndbufpt) + 1); } while (0)
#define PUT_SOUND_WORD(b) do { *(uae_u16 *)sndbufpt = b; sndbufpt = (uae_u16 *)(((uae_u8 *)sndbufpt) + 2); } while (0)
#define PUT_SOUND_BYTE_LEFT(b) PUT_SOUND_BYTE(b)
#define PUT_SOUND_WORD_LEFT(b) PUT_SOUND_WORD(b)
#define PUT_SOUND_BYTE_RIGHT(b) PUT_SOUND_BYTE(b)
#define PUT_SOUND_WORD_RIGHT(b) PUT_SOUND_WORD(b)
#define SOUND16_BASE_VAL 0
#define SOUND8_BASE_VAL 128

#define DEFAULT_SOUND_MAXB 8192
#define DEFAULT_SOUND_MINB 8192
#define DEFAULT_SOUND_BITS 16
#define DEFAULT_SOUND_FREQ 44100
#define HAVE_STEREO_SUPPORT
